(* File cont/backtrack.sml 

   An interpreter for micro-Icon, a language where an expression can
   produce more than one result.  This version has no variables.
   sestoft@dina.kvl.dk * 2003-03-29

   The evaluation of an expression may fail (producing no value), or
   may succeed once (producing one value), or may succeed several
   times (producing many values), or indeed, infinitely many times.

   The constant expression 5 succeeds once with the value 5.

   The range expression 

      1 to 3

   succeeds three times, with the values 1 2 3.

   The expression 
   
      write(1 to 3) 

   succeeds three times, printing 1 2 3.

   Evaluating 

      write(1 to 3)

   just prints 1.  To force it to produce (and print) all its success
   values, evaluate

      every(write(1 to 3))     

   The expression 

      (1 to 3) + (4 to 5) 

   succeeds 6 times, with values 5 6 6 7 7 8.

   The expression 

      (1 to 3) & (4 to 5) 

   succeeds 6 times, with the values 4 5 4 5 4 5.  It produces all
   values of the second expression for each value of the first
   expression.

   The expression 

      (1 to 3) | (4 to 5) 

   succeeds 5 times, with the values 1 2 3 4 5.  it produces all
   values of the first expression, followed by all values of the
   second expression.

   The expression 

      (1 to 3) ; (4 to 5) 

   succeeds 2 times, with the values 4 5.  It does not backtrack into
   the expression before the semicolon.

   The expression 

      (1 to 0) ; (4 to 5) 

   succeeds 2 times, with the values 4 5.

   The expression 

      (4 < 3) 

   fails.  Icon does not have Boolean values, instead a failing
   expression is considered false, and a successful expression is
   considered true.  The operators `&' (and) and `|' (or) agree with
   this interpretation, and behave like the sequential `&&' and `||'
   operators known from C/C++/Java/C#.

   The expression 

      (3 < 4) 

   succeeds once, with the value 4.

   The expression 

      (3 < (1 to 5)) 

   succeeds twice, with the values 4 5.

   The expression 

      if (1 to 3) then (11 to 12) else 22 

   succeeds twice, with 11 12.  Note that it does not backtrack into
   evaluating the condition (1 to 3).

   This language can be implemented by evaluating expressions with two
   continuations:

      * A normal (or success) continuation which is given the
        expression's result in case of success, and a failure
        continuation to use in case of future failures.

      * A failure continuation that is invoked in case of failure.
        This continuation may also be called the backtracking
        continuation, or a resumption (it resumes the execution of a
        multi-result expression).

   Thus in 
   
      write(1 to 3) & ...

   the success continuation of (1 to 3) will print 1, but that success
   continuation also takes as argument a failure continuation, so that
   if the evaluation of ... fails, then it will go back to (1 to 3) to
   produce 2 and then print 2, and so on.  After the value 3 has been
   produced, further attempts to backtrack should fail (by calling the
   failure continuation of the entire expression).  
*)

app load ["Int"];

datatype expr = 
    CstI of int                         (* i                     *)
  | Write of expr                       (* write(e)              *)
  | Prim of string * expr * expr        (* e1+e2, e1*e2, e1<e2   *)
  | And of expr * expr                  (* e1 & e2               *)
  | Or  of expr * expr                  (* e1 | e2               *)
  | Seq of expr * expr                  (* e1 ; e2               *)
  | If of expr * expr * expr            (* if e1 then e2 else e3 *)
  | Every of expr                       (* every(e)              *)
  | Fail                                (* &fail                 *)

(* Failure continuation and normal continuation *)

datatype answer =
    Success 
  | Failure of string;

type fcont = unit -> answer

type cont = int -> fcont -> answer

fun write i = (print (Int.toString i); print " ")

fun eval (e : expr) (cont : cont) (fcont : fcont) = 
    case e of
        CstI i       => cont i fcont
      | Write e      => 
            eval e (fn v => fn fcont1 => (write v; cont v fcont1)) fcont
      | Prim(ope, e1, e2) => 
            eval e1 (fn v1 => fn fcont1 =>
                     eval e2 (fn v2 => fn fcont2 => 
                              case ope of
                                  "+" => cont (v1+v2) fcont2 
                                | "*" => cont (v1*v2) fcont2
                                | "<" => 
                                      if v1<v2 then 
                                          cont v2 fcont2
                                      else
                                          fcont2 ()
                                | "to" => 
                                      let fun loop i = 
                                          if i <= v2 then 
                                              cont i (fn () => loop (i+1))
                                          else 
                                              fcont2 ()
                                      in loop v1 end
                                | _ => Failure "unknown primitive")
                          fcont1)
                    fcont
      | Or(e1, e2) => 
            eval e1 cont (fn () => eval e2 cont fcont)
      | And(e1, e2) => 
            eval e1 (fn _ => fn fcont1 => eval e2 cont fcont1) fcont
      | Seq(e1, e2) => 
            eval e1 (fn _ => fn _ => eval e2 cont fcont)
                    (fn () => eval e2 cont fcont)
      | If(e1, e2, e3) => 
            eval e1 (fn _ => fn _ => eval e2 cont fcont)
                    (fn () => eval e3 cont fcont)
      | Every e => 
            eval e (fn _ => fn fcont1 => fcont1 ()) fcont
      | Fail => fcont ()

fun run e = 
    eval e (fn v => fn _ => Success) (fn () => Failure "no result");

(* write (1 to 3) *)
val ex1 = Seq(Write (Prim("to", CstI 1, CstI 3)), Fail);

(* every (write (1 to 3)) *)
val ex2every = Every (Write (Prim("to", CstI 1, CstI 3)));

(* write (1 to 3) & fail *)
val ex2and = And(Write (Prim("to", CstI 1, CstI 3)), Fail);

(* if (write (1 to 3)) then fail else fail *)
val ex2if = If(Write (Prim("to", CstI 1, CstI 3)), Fail, Fail);

(* (write ((1 to 3) & (4 to 5))) & fail *)
val ex3and = And(Write(And(Prim("to", CstI 1, CstI 3), 
                           Prim("to", CstI 4, CstI 5))), Fail);

(* (write ((1 to 3) | (4 to 5))) & fail *)
val ex3or  = And(Write(Or(Prim("to", CstI 1, CstI 3), 
                          Prim("to", CstI 4, CstI 5))), Fail);

(* (write ((1 to 3) ; (4 to 5))) & fail *)
val ex3seq = And(Write(Seq(Prim("to", CstI 1, CstI 3), 
                           Prim("to", CstI 4, CstI 5))), Fail);

(* write ((1 to 3) & ((4 to 5) & 666)) *)
val ex4 = Write(And(Prim("to", CstI 1, CstI 3), 
                    And(Prim("to", CstI 4, CstI 5), CstI 666)));

(* (every (write (1 to 3))) & (4 to 5) *)
val ex6 = And(Every (Write (Prim("to", CstI 1, CstI 3))), 
              Prim("to", CstI 4, CstI 5));

(* every (write((1 to 3) + (4 to 5))) *)
val ex7 = Every(Write(Prim("+", Prim("to", CstI 1, CstI 3), 
                            Prim("to", CstI 4, CstI 5))));

(* write (4 < (1 to 10)) *)
val ex8 = Write(Prim("<", CstI 4, Prim("to", CstI 1, CstI 10)));

(* write (if (5 < 4) then 11 else 22) *)
val ex9 = Write(If(Prim("<", CstI 5, CstI 4), CstI 11, CstI 22));

(* every(write(if (1 to 3) then (11 to 12) else (21 to 22))) *)
val ex10 = Every(Write(If(Prim("to", CstI 1, CstI 3),
                          Prim("to", CstI 11, CstI 12),
                          Prim("to", CstI 21, CstI 22))));
                          
(* every (write ((1 to 3) to (4 to 5))) *)
val ex11 = Every(Write(Prim("to", Prim("to", CstI 1, CstI 3),
                                  Prim("to", CstI 4, CstI 5))));

(* every (write ((1 to 0) ; (4 to 5))) *)
val ex12 = Every(Write(Seq(Prim("to", CstI 1, CstI 0), 
                           Prim("to", CstI 4, CstI 5))));
